/*
 * Copyright (c) 2002-2005, Network Appliance, Inc. All rights reserved.
 *
 * This Software is licensed under one of the following licenses:
 * 
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see 
 *    http://www.opensource.org/licenses/cpl.php.
 * 
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 * 
 * 3) under the terms of the "GNU General Public License (GPL) Version 2" a
 *    copy of which is in the file LICENSE3.txt in the root directory. The
 *    license is also available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/gpl-license.php.
 * 
 * Licensee has the right to choose one of the above licenses.
 * 
 * Redistributions of source code must retain the above copyright
 * notice and one of the license notices.
 * 
 * Redistributions in binary form must reproduce both the above copyright
 * notice, one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/****************************************************************
 *
 * HEADER: kdat.h
 *
 * PURPOSE: defines the kernel DAT API
 *
 * Description: Header file for "kDAPL: Kernel Direct Access Programming
 *		Library, Version: 1.2"
 * Mapping rules:
 *      All global symbols are prepended with DAT_ or dat_
 *      All DAT objects have an 'api' tag which, such as 'ep' or 'lmr'
 *      The method table is in the provider definition structure.
 *
 *
 *
 ***************************************************************/

#ifndef _KDAT_H_
#define _KDAT_H_

#include <dat/kdat_config.h>

#include <dat/dat_platform_specific.h>

typedef enum dat_mem_type
{
	/* Shared between udat and kdat */
    DAT_MEM_TYPE_VIRTUAL        = 0x00,
    DAT_MEM_TYPE_LMR            = 0x01,
	/* kdat specific */
    DAT_MEM_TYPE_PHYSICAL       = 0x10,
    DAT_MEM_TYPE_PLATFORM       = 0x20,
    DAT_MEM_TYPE_IA             = 0x40,
    DAT_MEM_TYPE_BYPASS         = 0x80
} DAT_MEM_TYPE;

/* dat handle types */
typedef enum dat_handle_type
{
    DAT_HANDLE_TYPE_CR,
    DAT_HANDLE_TYPE_EP,
    DAT_HANDLE_TYPE_EVD,
    DAT_HANDLE_TYPE_IA,
    DAT_HANDLE_TYPE_LMR,
    DAT_HANDLE_TYPE_PSP,
    DAT_HANDLE_TYPE_PZ,
    DAT_HANDLE_TYPE_RMR,
    DAT_HANDLE_TYPE_RSP,
    DAT_HANDLE_TYPE_SRQ
} DAT_HANDLE_TYPE;

typedef enum dat_evd_param_mask
{
    DAT_EVD_FIELD_IA_HANDLE           = 0x01,
    DAT_EVD_FIELD_EVD_QLEN            = 0x02,
    DAT_EVD_FIELD_UPCALL_POLICY       = 0x04,
    DAT_EVD_FIELD_UPCALL              = 0x08,
    DAT_EVD_FIELD_EVD_FLAGS           = 0x10,
    DAT_EVD_FIELD_ALL                 = 0x1F
} DAT_EVD_PARAM_MASK;

typedef DAT_UINT64 DAT_PROVIDER_ATTR_MASK;

#include <dat/dat.h>

/* Upcall support */

typedef enum dat_upcall_policy
{
    DAT_UPCALL_DISABLE               = 0,   /* support no_upcalls           */
    DAT_UPCALL_SINGLE_INSTANCE       = 1,   /* support only one upcall      */
    DAT_UPCALL_MANY                  = 100, /* support multiple upcalls     */
    DAT_UPCALL_TEARDOWN              = -1   /* support no upcalls & return  *
					     * after all in progress        *
					     * UpCalls return               */
} DAT_UPCALL_POLICY;

typedef void (*DAT_UPCALL_FUNC)(
	DAT_PVOID,		/* instance_data        */
	const DAT_EVENT *,	/* event                */
	DAT_BOOLEAN);		/* more_events          */

typedef struct dat_upcall_object
{
    DAT_PVOID                   instance_data;
    DAT_UPCALL_FUNC             upcall_func;
} DAT_UPCALL_OBJECT;

/* Define NULL upcall */

#define DAT_UPCALL_NULL \
	(DAT_UPCALL_OBJECT) { (DAT_PVOID) NULL, \
        (DAT_UPCALL_FUNC) NULL}

#define DAT_UPCALL_SAME		((DAT_UPCALL_OBJECT *) NULL)

struct dat_evd_param
{
    DAT_IA_HANDLE               ia_handle;
    DAT_COUNT                   evd_qlen;
    DAT_UPCALL_POLICY           upcall_policy;
    DAT_UPCALL_OBJECT           upcall;
    DAT_EVD_FLAGS               evd_flags;
};

/*
 * Memory types
 *
 * Specifying memory type for LMR create. A Consumer must use a single
 * value when registering memory. The union of any of these
 * flags is used in the Provider parameters to indicate what memory
 * type Provider supports for LMR memory creation.
 */

/* memory data types */
#define DAT_MEM_OPT DAT_MEM_OPTIMIZE_FLAGS

typedef enum dat_mem_optimize_flags
{
    DAT_MEM_OPTIMIZE_DONT_CARE       = 0x00,
    DAT_MEM_OPTIMIZE_IA              = 0x01,
    DAT_MEM_OPTIMIZE_MIN_EXPOSURE    = 0x02,
    DAT_MEM_OPTIMIZE_EXACT_EXPOSURE  = 0x04
} DAT_MEM_OPTIMIZE_FLAGS;


typedef union dat_region_description
{
    DAT_PVOID                   for_va;
    DAT_LMR_HANDLE              for_lmr_handle;
    void *                      for_pointer;	/* For kdapl only */
    void *                      for_array;	/* For kdapl only */
    DAT_PADDR                   for_pa;		/* For kdapl only */
} DAT_REGION_DESCRIPTION;

/* LMR Arguments */
struct dat_lmr_param
{
    DAT_IA_HANDLE               ia_handle;
    DAT_MEM_TYPE                mem_type;
    DAT_REGION_DESCRIPTION      region_desc;
    DAT_VLEN                    length;
    DAT_PZ_HANDLE               pz_handle;
    DAT_MEM_PRIV_FLAGS          mem_priv;
    DAT_LMR_CONTEXT             lmr_context;
    DAT_RMR_CONTEXT             rmr_context;
    DAT_VLEN                    registered_size;
    DAT_VADDR                   registered_address;
};

#include <dat/kdat_vendor_specific.h>

/* General Provider attributes. kdat specific. */
/* Provider should support merging of all event stream types. Provider
 * attribute specify support for merging different event stream types.
 * It is a 2D binary matrix where each row and column represents an event
 * stream type. Each binary entry is 1 if the event streams of its raw
 * and column can fed the same EVD, and 0 otherwise. The order of event
 * streams in row and column is the same as in the definition of
 * DAT_EVD_FLAGS: index 0 - Software Event, 1- Connection Request,
 * 2 - DTO Completion, 3 - Connection event, 4 - RMR Bind Completion,
 * 5 - Asynchronous event. By definition each diagonal entry is 1.
 * Consumer allocates an array for it and passes it IN as a pointer
 * for the array that Provider fills. Provider must fill the array
 * that Consumer passes.
 */

struct dat_provider_attr
{
    char                        provider_name[DAT_NAME_MAX_LENGTH];
    DAT_UINT32                  provider_version_major;
    DAT_UINT32                  provider_version_minor;
    DAT_UINT32                  dapl_version_major;
    DAT_UINT32                  dapl_version_minor;
    DAT_MEM_TYPE                lmr_mem_types_supported;
    DAT_IOV_OWNERSHIP           iov_ownership_on_return;
    DAT_QOS                     dat_qos_supported;
    DAT_COMPLETION_FLAGS        completion_flags_supported;
    DAT_BOOLEAN                 is_thread_safe;
    DAT_COUNT                   max_private_data_size;
    DAT_BOOLEAN                 supports_multipath;
    DAT_EP_CREATOR_FOR_PSP      ep_creator;
    DAT_UPCALL_POLICY           upcall_policy;
    DAT_UINT32                  optimal_buffer_alignment;
    const DAT_BOOLEAN           evd_stream_merging_supported[6][6];
    DAT_BOOLEAN                 srq_supported;
    DAT_COUNT                   srq_watermarks_supported;
    DAT_BOOLEAN                 srq_ep_pz_difference_supported;
    DAT_COUNT                   srq_info_supported;
    DAT_COUNT                   ep_recv_info_supported;
    DAT_BOOLEAN                 lmr_sync_req;
    DAT_BOOLEAN                 dto_async_return_guaranteed;
    DAT_BOOLEAN                 rdma_write_for_rdma_read_req;
    DAT_COUNT                   num_provider_specific_attr;
    DAT_NAMED_ATTR *            provider_specific_attr;
};


#define DAT_PROVIDER_FIELD_PROVIDER_NAME                 UINT64_C(0x0000001)
#define DAT_PROVIDER_FIELD_PROVIDER_VERSION_MAJOR        UINT64_C(0x0000002)
#define DAT_PROVIDER_FIELD_PROVIDER_VERSION_MINOR        UINT64_C(0x0000004)
#define DAT_PROVIDER_FIELD_DAPL_VERSION_MAJOR            UINT64_C(0x0000008)
#define DAT_PROVIDER_FIELD_DAPL_VERSION_MINOR            UINT64_C(0x0000010)
#define DAT_PROVIDER_FIELD_LMR_MEM_TYPE_SUPPORTED        UINT64_C(0x0000020)
#define DAT_PROVIDER_FIELD_IOV_OWNERSHIP                 UINT64_C(0x0000040)
#define DAT_PROVIDER_FIELD_DAT_QOS_SUPPORTED             UINT64_C(0x0000080)
#define DAT_PROVIDER_FIELD_COMPLETION_FLAGS_SUPPORTED    UINT64_C(0x000100)
#define DAT_PROVIDER_FIELD_IS_THREAD_SAFE                UINT64_C(0x0000200)
#define DAT_PROVIDER_FIELD_MAX_PRIVATE_DATA_SIZE         UINT64_C(0x0000400)
#define DAT_PROVIDER_FIELD_SUPPORTS_MULTIPATH            UINT64_C(0x0000800)
#define DAT_PROVIDER_FIELD_EP_CREATOR                    UINT64_C(0x0001000)
#define DAT_PROVIDER_FIELD_UPCALL_POLICY                 UINT64_C(0x0002000)
#define DAT_PROVIDER_FIELD_OPTIMAL_BUFFER_ALIGNMENT      UINT64_C(0x0004000)
#define DAT_PROVIDER_FIELD_EVD_STREAM_MERGING_SUPPORTED  UINT64_C(0x0008000)
#define DAT_PROVIDER_FIELD_SRQ_SUPPORTED                 UINT64_C(0x0010000)
#define DAT_PROVIDER_FIELD_SRQ_WATERMARKS_SUPPORTED      UINT64_C(0x0020000)
#define DAT_PROVIDER_FIELD_SRQ_EP_PZ_DIFFERENCE_SUPPORTED UINT64_C(0x0040000)
#define DAT_PROVIDER_FIELD_SRQ_INFO_SUPPORTED            UINT64_C(0x0080000)
#define DAT_PROVIDER_FIELD_EP_RECV_INFO_SUPPORTED        UINT64_C(0x0100000)
#define DAT_PROVIDER_FIELD_LMR_SYNC_REQ                  UINT64_C(0x0200000)
#define DAT_PROVIDER_FIELD_DTO_ASYNC_RETURN_GUARANTEED   UINT64_C(0x0400000)
#define DAT_PROVIDER_FIELD_RDMA_WRITE_FOR_RDMA_READ_REQ  UINT64_C(0x0800000)
#define DAT_PROVIDER_FIELD_NUM_PROVIDER_SPECIFIC_ATTR    UINT64_C(0x1000000)
#define DAT_PROVIDER_FIELD_PROVIDER_SPECIFIC_ATTR        UINT64_C(0x2000000)

#define DAT_PROVIDER_FIELD_ALL                           UINT64_C(0x3FFFFFF)
#define DAT_PROVIDER_FIELD_NONE				 UINT64_C(0x0)

/**************************************************************/
/*
 * Kernel DAT function call definitions,
 */

extern DAT_RETURN dat_lmr_kcreate (
	IN      DAT_IA_HANDLE,			/* ia_handle            */
	IN      DAT_MEM_TYPE,			/* mem_type             */
	IN      DAT_REGION_DESCRIPTION,		/* region_description   */
	IN      DAT_VLEN,			/* length               */
	IN      DAT_PZ_HANDLE,			/* pz_handle            */
	IN      DAT_MEM_PRIV_FLAGS,		/* privileges           */
	IN      DAT_MEM_OPTIMIZE_FLAGS,		/* mem_optimization     */
	OUT     DAT_LMR_HANDLE *,		/* lmr_handle           */
	OUT     DAT_LMR_CONTEXT *,		/* lmr_context          */
	OUT     DAT_RMR_CONTEXT *,		/* rmr_context          */
	OUT     DAT_VLEN *,			/* registered_length    */
	OUT     DAT_VADDR * );			/* registered_address   */

extern DAT_RETURN dat_ia_memtype_hint (
	IN      DAT_IA_HANDLE,			/* ia_handle            */
	IN      DAT_MEM_TYPE,			/* mem_type             */
	IN      DAT_VLEN,			/* length               */
	IN      DAT_MEM_OPTIMIZE_FLAGS,		/* mem_optimization     */
	OUT     DAT_VLEN *,			/* preferred_length     */
	OUT     DAT_VADDR * );			/* preferred_alignment  */

extern DAT_RETURN dat_evd_kcreate (
	IN      DAT_IA_HANDLE,			/* ia_handle            */
	IN      DAT_COUNT,			/* evd_min_qlen         */
	IN      DAT_UPCALL_POLICY,		/* upcall_policy        */
	IN      const DAT_UPCALL_OBJECT *,	/* upcall               */
	IN      DAT_EVD_FLAGS,			/* evd_flags            */
	OUT     DAT_EVD_HANDLE * );		/* evd_handle           */

extern DAT_RETURN dat_evd_modify_upcall (
	IN      DAT_EVD_HANDLE,			/* evd_handle           */
	IN      DAT_UPCALL_POLICY,		/* upcall_policy        */
	IN      const DAT_UPCALL_OBJECT * );	/* upcall               */

#endif /* _KDAT_H_ */
